﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// 
  /// </summary>
  [StorableClass]
  [Item("ContinuousPointCertaintyWeightCalculator", "")]
  public class ContinuousPointCertaintyWeightCalculator : DiscriminantClassificationWeightCalculator {

    public ContinuousPointCertaintyWeightCalculator()
      : base() {
    }

    [StorableConstructor]
    protected ContinuousPointCertaintyWeightCalculator(bool deserializing) : base(deserializing) { }
    protected ContinuousPointCertaintyWeightCalculator(ContinuousPointCertaintyWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ContinuousPointCertaintyWeightCalculator(this, cloner);
    }

    protected override IEnumerable<double> DiscriminantCalculateWeights(IEnumerable<IDiscriminantFunctionClassificationSolution> discriminantSolutions) {
      List<double> weights = new List<double>();
      IClassificationProblemData problemData = discriminantSolutions.ElementAt(0).ProblemData;
      IEnumerable<double> targetValues;
      IEnumerator<double> trainingValues;

      //only works for binary classification
      if (!problemData.ClassValues.Count().Equals(2))
        return Enumerable.Repeat<double>(1, discriminantSolutions.Count());

      double maxClass = problemData.ClassValues.Max();
      double minClass = problemData.ClassValues.Min();
      double halfDistanceBetweenClasses = (maxClass - minClass) / 2;

      foreach (var solution in discriminantSolutions) {
        problemData = solution.ProblemData;
        targetValues = GetValues(problemData.Dataset.GetDoubleValues(problemData.TargetVariable).ToList(), problemData.TrainingIndizes);
        trainingValues = targetValues.GetEnumerator();

        IEnumerator<double> estimatedTrainingVal = solution.EstimatedTrainingValues.GetEnumerator();
        IEnumerator<double> estimatedTrainingClassVal = solution.EstimatedTrainingClassValues.GetEnumerator();

        double curWeight = 0.0;
        while (estimatedTrainingVal.MoveNext() && estimatedTrainingClassVal.MoveNext() && trainingValues.MoveNext()) {
          if (trainingValues.Current.Equals(maxClass)) {
            if (estimatedTrainingVal.Current >= maxClass)
              curWeight += 1.0;
            else {
              double distanceToPoint = Math.Abs(estimatedTrainingVal.Current - maxClass);
              curWeight += Math.Max(1.0 - (1.0 / halfDistanceBetweenClasses) * distanceToPoint, -1.0);
            }
          } else if (trainingValues.Current.Equals(minClass)) {
            if (estimatedTrainingVal.Current <= minClass)
              curWeight += 1.0;
            else {
              double distanceToPoint = Math.Abs(estimatedTrainingVal.Current - minClass);
              curWeight += Math.Max(1.0 - (1.0 / halfDistanceBetweenClasses) * distanceToPoint, -1.0);
            }
          }
        }
        // normalize the weight (otherwise a model with a bigger training partition would probably be better)
        weights.Add(curWeight / targetValues.Count());
      }
      return weights;
    }
  }
}
